# Copyright 2015 Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fastfood core module."""

from __future__ import print_function
from __future__ import unicode_literals

import datetime
import errno
import json
import logging
import os
import shutil

from fastfood import book
from fastfood import pack
from fastfood import templating
from fastfood import utils

LOG = logging.getLogger(__name__)


def _determine_selected_stencil(stencil_set, stencil_definition):
    """Determine appropriate stencil name for stencil definition.

    Given a fastfood.json stencil definition with a stencil set, figure out
    what the name of the stencil within the set should be, or use the default
    """
    if 'stencil' not in stencil_definition:
        selected_stencil_name = stencil_set.manifest.get('default_stencil')
    else:
        selected_stencil_name = stencil_definition.get('stencil')

    if not selected_stencil_name:
        raise ValueError("No stencil name, within stencil set %s, specified."
                         % stencil_definition['name'])

    return selected_stencil_name


def _build_template_map(cookbook, cookbook_name, stencil):
    """Build a map of variables for this generated cookbook and stencil.

    Get template variables from stencil option values, adding the default ones
    like cookbook and cookbook year.
    """
    template_map = {
        'cookbook': {"name": cookbook_name},
        'options': stencil['options']
    }

    # Cookbooks may not yet have metadata, so we pass an empty dict if so
    try:
        template_map['cookbook'] = cookbook.metadata.to_dict().copy()
    except ValueError:
        # ValueError may be returned if this cookbook does not yet have any
        # metadata.rb written by a stencil. This is okay, as everyone should
        # be using the base stencil first, and then we'll try to call
        # cookbook.metadata again in this method later down.
        pass

    template_map['cookbook']['year'] = datetime.datetime.now().year
    return template_map


def _render_binaries(files, written_files):
    """Write binary contents from filetable into files.

    Using filetable for the input files, and the list of files, render
    all the templates into actual files on disk, forcing to overwrite the file
    as appropriate, and using the given open mode for the file.
    """
    for source_path, target_path in files.items():
        needdir = os.path.dirname(target_path)
        assert needdir, "Target should have valid parent dir"
        try:
            os.makedirs(needdir)
        except OSError as err:
            if err.errno != errno.EEXIST:
                raise

        if os.path.isfile(target_path):
            if target_path in written_files:
                LOG.warning("Previous stencil has already written file %s.",
                            target_path)
            else:
                print("Skipping existing file %s" % target_path)
                LOG.info("Skipping existing file %s", target_path)
                continue

        print("Writing rendered file %s" % target_path)
        LOG.info("Writing rendered file %s", target_path)

        shutil.copy(source_path, target_path)
        if os.path.exists(target_path):
            written_files.append(target_path)


def _render_templates(files, filetable, written_files, force, open_mode='w'):
    """Write template contents from filetable into files.

    Using filetable for the rendered templates, and the list of files, render
    all the templates into actual files on disk, forcing to overwrite the file
    as appropriate, and using the given open mode for the file.
    """
    for tpl_path, content in filetable:
        target_path = files[tpl_path]
        needdir = os.path.dirname(target_path)
        assert needdir, "Target should have valid parent dir"
        try:
            os.makedirs(needdir)
        except OSError as err:
            if err.errno != errno.EEXIST:
                raise

        if os.path.isfile(target_path):
            if force:
                LOG.warning("Forcing overwrite of existing file %s.",
                            target_path)
            elif target_path in written_files:
                LOG.warning("Previous stencil has already written file %s.",
                            target_path)
            else:
                print("Skipping existing file %s" % target_path)
                LOG.info("Skipping existing file %s", target_path)
                continue

        with open(target_path, open_mode) as newfile:
            print("Writing rendered file %s" % target_path)
            LOG.info("Writing rendered file %s", target_path)
            newfile.write(content)
            written_files.append(target_path)


def build_cookbook(build_config, templatepack_path,
                   cookbooks_home, force=False):
    """Build a cookbook from a fastfood.json file.

    Can build on an existing cookbook, otherwise this will
    create a new cookbook for you based on your templatepack.
    """
    with open(build_config) as cfg:
        cfg = json.load(cfg)

    cookbook_name = cfg['name']
    template_pack = pack.TemplatePack(templatepack_path)

    written_files = []
    cookbook = create_new_cookbook(cookbook_name, cookbooks_home)

    for stencil_definition in cfg['stencils']:

        selected_stencil_set_name = stencil_definition.get('stencil_set')
        stencil_set = template_pack.load_stencil_set(selected_stencil_set_name)

        selected_stencil_name = _determine_selected_stencil(
            stencil_set,
            stencil_definition
        )

        stencil = stencil_set.get_stencil(selected_stencil_name,
                                          **stencil_definition)

        updated_cookbook = process_stencil(
            cookbook,
            cookbook_name,  # in case no metadata.rb yet
            template_pack,
            force,
            stencil_set,
            stencil,
            written_files
        )

    return written_files, updated_cookbook


def process_stencil(cookbook, cookbook_name, template_pack,
                    force_argument, stencil_set, stencil, written_files):
    """Process the stencil requested, writing any missing files as needed.

    The stencil named 'stencilset_name' should be one of
    templatepack's stencils.
    """
    # force can be passed on the command line or forced in a stencil's options
    force = force_argument or stencil['options'].get('force', False)

    stencil['files'] = stencil.get('files') or {}
    files = {
        # files.keys() are template paths, files.values() are target paths
        # {path to template: rendered target path, ... }
        os.path.join(stencil_set.path, tpl): os.path.join(cookbook.path, tgt)
        for tgt, tpl in stencil['files'].items()
    }

    stencil['partials'] = stencil.get('partials') or {}
    partials = {
        # files.keys() are template paths, files.values() are target paths
        # {path to template: rendered target path, ... }
        os.path.join(stencil_set.path, tpl): os.path.join(cookbook.path, tgt)
        for tgt, tpl in stencil['partials'].items()
    }

    stencil['binaries'] = stencil.get('binaries') or {}
    binaries = {
        # files.keys() are binary paths, files.values() are target paths
        # {path to binary: rendered target path, ... }
        os.path.join(stencil_set.path, tpl): os.path.join(cookbook.path, tgt)
        for tgt, tpl in stencil['binaries'].items()
    }

    template_map = _build_template_map(cookbook, cookbook_name, stencil)

    filetable = templating.render_templates(*files.keys(), **template_map)
    _render_templates(files, filetable, written_files, force)

    parttable = templating.render_templates(*partials.keys(), **template_map)
    _render_templates(partials, parttable, written_files, force, open_mode='a')

    # no templating needed for binaries, just pass off to the copy method
    _render_binaries(binaries, written_files)

    # merge metadata.rb dependencies
    stencil_metadata_deps = {'depends': stencil.get('dependencies', {})}
    stencil_metadata = book.MetadataRb.from_dict(stencil_metadata_deps)
    cookbook.metadata.merge(stencil_metadata)

    # merge Berksfile dependencies
    stencil_berks_deps = {'cookbook': stencil.get('berks_dependencies', {})}
    stencil_berks = book.Berksfile.from_dict(stencil_berks_deps)
    cookbook.berksfile.merge(stencil_berks)

    return cookbook


def create_new_cookbook(cookbook_name, cookbooks_home):
    """Create a new cookbook.

    :param cookbook_name: Name of the new cookbook.
    :param cookbooks_home: Target dir for new cookbook.
    """
    cookbooks_home = utils.normalize_path(cookbooks_home)

    if not os.path.exists(cookbooks_home):
        raise ValueError("Target cookbook dir %s does not exist."
                         % os.path.relpath(cookbooks_home))

    target_dir = os.path.join(cookbooks_home, cookbook_name)
    LOG.debug("Creating dir -> %s", target_dir)
    try:
        os.makedirs(target_dir)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise
        else:
            LOG.info("Skipping existing directory %s", target_dir)

    cookbook_path = os.path.join(cookbooks_home, cookbook_name)
    cookbook = book.CookBook(cookbook_path)

    return cookbook



# Copyright 2015 Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=invalid-name

"""Fastfood Exceptions."""

import re

# python 2 vs. 3 string types
try:
    basestring
except NameError:
    basestring = str

_SPLITCASE_RE = re.compile(r'[A-Z][^A-Z]*')


class FastfoodError(Exception):

    """Base class for all exceptions raised by fastfood."""


class FastfoodStencilSetNotListed(FastfoodError):

    """Stencil set specified was not listed in the templatepack manifest."""


class FastfoodStencilSetInvalidPath(FastfoodError):

    """Specified path to stencil set does not exist."""


class FastfoodStencilSetMissingManifest(FastfoodError):

    """Stencil set is missing a manifest.json file."""


class FastfoodTemplatePackAttributeError(AttributeError, FastfoodError):

    """Invalid stencilset request from TemplatePack."""


def get_friendly_title(err):
    """Turn class, instance, or name (str) into an eyeball-friendly title.

    E.g. FastfoodStencilSetNotListed --> 'Stencil Set Not Listed'
    """
    if isinstance(err, basestring):
        string = err
    else:
        try:
            string = err.__name__
        except AttributeError:
            string = err.__class__.__name__
    split = _SPLITCASE_RE.findall(string)
    if not split:
        split.append(string)
    if len(split) > 1 and split[0] == 'Fastfood':
        split.pop(0)

    return " ".join(split)



# Copyright 2015 Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fastfood Chef Cookbook manager."""
from __future__ import print_function

import os

from fastfood import utils


class CookBook(object):

    """Chef Cookbook object.

    Understands metadata.rb, Berksfile and how to parse them.
    """

    def __init__(self, path):
        """Initialize CookBook wrapper at 'path'."""
        self.path = utils.normalize_path(path)
        self._metadata = None
        if not os.path.isdir(path):
            raise ValueError("Cookbook dir %s does not exist."
                             % self.path)
        self._berksfile = None

    @property
    def name(self):
        """Cookbook name property."""
        try:
            return self.metadata.to_dict()['name']
        except KeyError:
            raise LookupError("%s is missing 'name' attribute'."
                              % self.metadata)

    @property
    def metadata(self):
        """Return dict representation of this cookbook's metadata.rb ."""
        self.metadata_path = os.path.join(self.path, 'metadata.rb')
        if not os.path.isfile(self.metadata_path):
            raise ValueError("Cookbook needs metadata.rb, %s"
                             % self.metadata_path)

        if not self._metadata:
            self._metadata = MetadataRb(open(self.metadata_path, 'r+'))

        return self._metadata

    @property
    def berksfile(self):
        """Return this cookbook's Berksfile instance."""
        self.berks_path = os.path.join(self.path, 'Berksfile')
        if not self._berksfile:
            if not os.path.isfile(self.berks_path):
                raise ValueError("No Berksfile found at %s"
                                 % self.berks_path)
            self._berksfile = Berksfile(open(self.berks_path, 'r+'))
        return self._berksfile


class MetadataRb(utils.FileWrapper):

    """Wrapper for a metadata.rb file."""

    @classmethod
    def from_dict(cls, dictionary):
        """Create a MetadataRb instance from a dict."""
        cookbooks = set()
        # put these in order
        groups = [cookbooks]

        for key, val in dictionary.items():
            if key == 'depends':
                cookbooks.update({cls.depends_statement(cbn, meta)
                                  for cbn, meta in val.items()})

        body = ''
        for group in groups:
            if group:
                body += '\n'
            body += '\n'.join(group)
        return cls.from_string(body)

    @staticmethod
    def depends_statement(cookbook_name, metadata=None):
        """Return a valid Ruby 'depends' statement for the metadata.rb file."""
        line = "depends '%s'" % cookbook_name
        if metadata:
            if not isinstance(metadata, dict):
                raise TypeError("Stencil dependency options for %s "
                                "should be a dict of options, not %s."
                                % (cookbook_name, metadata))
            if metadata:
                line = "%s '%s'" % (line, "', '".join(metadata))
        return line

    def to_dict(self):
        """Return a dictionary representation of this metadata.rb file."""
        return self.parse()

    def parse(self):
        """Parse the metadata.rb into a dict."""
        data = utils.ruby_lines(self.readlines())
        data = [tuple(j.strip() for j in line.split(None, 1))
                for line in data]
        depends = {}
        for line in data:
            if not len(line) == 2:
                continue
            key, value = line
            if key == 'depends':
                value = value.split(',')
                lib = utils.ruby_strip(value[0])
                detail = [utils.ruby_strip(j) for j in value[1:]]
                depends[lib] = detail
        datamap = {key: utils.ruby_strip(val) for key, val in data}
        if depends:
            datamap['depends'] = depends
        self.seek(0)
        return datamap

    def merge(self, other):
        """Add requirements from 'other' metadata.rb into this one."""
        if not isinstance(other, MetadataRb):
            raise TypeError("MetadataRb to merge should be a 'MetadataRb' "
                            "instance, not %s.", type(other))
        current = self.to_dict()
        new = other.to_dict()

        # compare and gather cookbook dependencies
        meta_writelines = ['%s\n' % self.depends_statement(cbn, meta)
                           for cbn, meta in new.get('depends', {}).items()
                           if cbn not in current.get('depends', {})]

        self.write_statements(meta_writelines)
        return self.to_dict()


class Berksfile(utils.FileWrapper):

    """Wrapper for a Berksfile."""

    berks_options = [
        'branch',
        'git',
        'path',
        'ref',
        'revision',
        'tag',
    ]

    def to_dict(self):
        """Return a dictionary representation of this Berksfile."""
        return self.parse()

    def parse(self):
        """Parse this Berksfile into a dict."""
        self.flush()
        self.seek(0)
        data = utils.ruby_lines(self.readlines())
        data = [tuple(j.strip() for j in line.split(None, 1))
                for line in data]
        datamap = {}
        for line in data:
            if len(line) == 1:
                datamap[line[0]] = True
            elif len(line) == 2:
                key, value = line
                if key == 'cookbook':
                    datamap.setdefault('cookbook', {})
                    value = [utils.ruby_strip(v) for v in value.split(',')]
                    lib, detail = value[0], value[1:]
                    datamap['cookbook'].setdefault(lib, {})
                    # if there is additional dependency data but its
                    # not the ruby hash, its the version constraint
                    if detail and not any("".join(detail).startswith(o)
                                          for o in self.berks_options):
                        constraint, detail = detail[0], detail[1:]
                        datamap['cookbook'][lib]['constraint'] = constraint
                    if detail:
                        for deet in detail:
                            opt, val = [
                                utils.ruby_strip(i)
                                for i in deet.split(':', 1)
                            ]
                            if not any(opt == o for o in self.berks_options):
                                raise ValueError(
                                    "Cookbook detail '%s' does not specify "
                                    "one of '%s'" % (opt, self.berks_options))
                            else:
                                datamap['cookbook'][lib][opt.strip(':')] = (
                                    utils.ruby_strip(val))
                elif key == 'source':
                    datamap.setdefault(key, [])
                    datamap[key].append(utils.ruby_strip(value))
                elif key:
                    datamap[key] = utils.ruby_strip(value)
        self.seek(0)
        return datamap

    @classmethod
    def from_dict(cls, dictionary):
        """Create a Berksfile instance from a dict."""
        cookbooks = set()
        sources = set()
        other = set()
        # put these in order
        groups = [sources, cookbooks, other]

        for key, val in dictionary.items():
            if key == 'cookbook':
                cookbooks.update({cls.cookbook_statement(cbn, meta)
                                  for cbn, meta in val.items()})
            elif key == 'source':
                sources.update({"source '%s'" % src for src in val})
            elif key == 'metadata':
                other.add('metadata')

        body = ''
        for group in groups:
            if group:
                body += '\n'
            body += '\n'.join(group)
        return cls.from_string(body)

    @staticmethod
    def cookbook_statement(cookbook_name, metadata=None):
        """Return a valid Ruby 'cookbook' statement for the Berksfile."""
        line = "cookbook '%s'" % cookbook_name
        if metadata:
            if not isinstance(metadata, dict):
                raise TypeError("Berksfile dependency hash for %s "
                                "should be a dict of options, not %s."
                                % (cookbook_name, metadata))
            # not like the others...
            if 'constraint' in metadata:
                line += ", '%s'" % metadata.pop('constraint')
            for opt, spec in metadata.items():
                line += ", %s: '%s'" % (opt, spec)
        return line

    def merge(self, other):
        """Add requirements from 'other' Berksfile into this one."""
        if not isinstance(other, Berksfile):
            raise TypeError("Berksfile to merge should be a 'Berksfile' "
                            "instance, not %s.", type(other))
        current = self.to_dict()
        new = other.to_dict()

        # compare and gather cookbook dependencies
        berks_writelines = ['%s\n' % self.cookbook_statement(cbn, meta)
                            for cbn, meta in new.get('cookbook', {}).items()
                            if cbn not in current.get('cookbook', {})]

        # compare and gather 'source' requirements
        berks_writelines.extend(["source '%s'\n" % src for src
                                 in new.get('source', [])
                                 if src not in current.get('source', [])])

        self.write_statements(berks_writelines)
        return self.to_dict()



# Copyright 2015 Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fastfood package attributes and metadata."""

__all__ = (
    '__title__',
    '__summary__',
    '__author__',
    '__email__',
    '__version__',
    '__license__',
    '__copyright__',
    '__url__',
)

__title__ = 'fastfood'
__summary__ = 'Chef Cookbook Wizardry'
__author__ = 'Rackers'
__email__ = 'samuel.stavinoha@rackspace.com'
__version__ = '1.1.0'
__keywords__ = ['chef', 'cookbook', 'templating', 'generation']
__license__ = 'Apache License, Version 2.0'
__copyright__ = 'Copyright Rackspace US, Inc. 2015'
__url__ = 'https://github.com/rackerlabs/fastfood'



#!/usr/bin/env python

# Copyright 2015 Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=invalid-name


"""Fastfood packaging and installation."""

import os
from setuptools import setup, find_packages

src_dir = os.path.dirname(os.path.realpath(__file__))

about = {}
with open(os.path.join(src_dir, 'fastfood', '__about__.py')) as abt:
    exec(abt.read(), about)


# README.rst is for fastfood's PyPI page
# pandoc --from=markdown_github --to=rst README.md --output=README.rst
with open(os.path.join(src_dir, 'README.rst')) as rdme:
    LONG_DESCRIPTION = rdme.read()


INSTALL_REQUIRES = [
    'Jinja2==2.7.3',
]


TESTS_REQUIRE = [
    'coverage==3.7.1',
    'flake8==2.4.1',
    'flake8-docstrings==0.2.1',
    'mock==1.2.0',
    'nose==1.3.7',
    'nose-ignore-docstring==0.2',
    'pylint==1.4.4',
    'vcrpy==1.6.1',
]


CLASSIFIERS = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Topic :: Software Development',
    'Topic :: Software Development :: Code Generators',
    'Development Status :: 4 - Beta',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
]


ENTRY_POINTS = {
    'console_scripts': ['fastfood=fastfood.shell:main'],
}


package_attributes = {
    'author': about['__author__'],
    'author_email': about['__email__'],
    'classifiers': CLASSIFIERS,
    'entry_points': ENTRY_POINTS,
    'name': about['__title__'],
    'description': about['__summary__'],
    'install_requires': INSTALL_REQUIRES,
    'keywords': ' '.join(about['__keywords__']),
    'license': about['__license__'],
    'long_description': LONG_DESCRIPTION,
    'packages': find_packages(exclude=['tests']),
    'test_suite': 'tests',
    'tests_require': TESTS_REQUIRE,
    'url': about['__url__'],
    'version': about['__version__'],
}


setup(**package_attributes)



# Copyright 2015 Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=invalid-name


"""Fastfood utils."""
from __future__ import print_function

import copy
import os
try:
    from StringIO import StringIO
except ImportError:
    # Python 3
    from io import StringIO

# python 2 vs. 3 string types
try:
    basestring
except NameError:
    basestring = str


def normalize_path(path):
    """Normalize and return absolute path.

    Expand user symbols like ~ and resolve relative paths.
    """
    return os.path.abspath(os.path.expanduser(os.path.normpath(path)))


def ruby_strip(chars):
    """Strip whitespace and any quotes."""
    return chars.strip(' "\'')


def ruby_lines(text):
    """Tidy up lines from a file, honor # comments.

    Does not honor ruby block comments (yet).
    """
    if isinstance(text, basestring):
        text = text.splitlines()
    elif not isinstance(text, list):
        raise TypeError("text should be a list or a string, not %s"
                        % type(text))
    return [l.strip() for l in text if l.strip() and not
            l.strip().startswith('#')]


def deepupdate(original, update, levels=5):
    """Update, like dict.update, but deeper.

    Update 'original' from dict/iterable 'update'.

    I.e., it recurses on dicts 'levels' times if necessary.

    A standard dict.update is levels=0
    """
    if not isinstance(update, dict):
        update = dict(update)
    if not levels > 0:
        original.update(update)
    else:
        for key, val in update.items():
            if isinstance(original.get(key), dict):
                # might need a force=True to override this
                if not isinstance(val, dict):
                    raise TypeError("Trying to update dict %s with "
                                    "non-dict %s" % (original[key], val))
                deepupdate(original[key], val, levels=levels-1)
            else:
                original.update({key: val})


class FileWrapper(object):

    """Helps wrap ruby files, usually.

    Like metadata.rb and Berksfile.
    """

    def __init__(self, stream):
        """Initialize the wrapper with a file-like object."""
        self.stream = stream

    @classmethod
    def from_string(cls, contents):
        """Initialize class with a string."""
        stream = StringIO(contents)
        return cls(stream)

    def __str__(self):
        """String representation of the object."""
        return '%s(%s)' % (type(self).__name__, self.stream)

    def __repr__(self):
        """Canonical string representation of the object."""
        return ('<%s [%r] at %s>' % (type(self).__name__,
                                     self.stream,
                                     hex(id(self))))

    def __getattr__(self, attr):
        """Shortcut to the file-like object's methods."""
        try:
            return getattr(self.stream, attr)
        except AttributeError:
            raise AttributeError("'%s' object has no attribute '%s'"
                                 % (type(self).__name__, attr))

    def write_statements(self, statements):
        """Insert the statements into the file neatly.

        Ex:

        statements = ["good  'dog'", "good 'cat'", "bad  'rat'", "fat 'emu'"]

        # stream.txt ... animals = FileWrapper(open('stream.txt'))
        good 'cow'
        nice 'man'
        bad 'news'

        animals.write_statements(statements)

        # stream.txt
        good 'cow'
        good 'dog'
        good 'cat'
        nice 'man'
        bad 'news'
        bad 'rat'
        fat 'emu'
        """
        self.seek(0)
        original_content_lines = self.readlines()
        new_content_lines = copy.copy(original_content_lines)
        # ignore blanks and sort statements to be written
        statements = sorted([stmnt for stmnt in statements if stmnt])

        # find all the insert points for each statement
        uniqs = {stmnt.split(None, 1)[0] for stmnt in statements}
        insert_locations = {}
        for line in reversed(original_content_lines):
            if not uniqs:
                break
            if not line:
                continue
            for word in uniqs.copy():
                if line.startswith(word):
                    index = original_content_lines.index(line) + 1
                    insert_locations[word] = index
                    uniqs.remove(word)

        for statement in statements:
            print("writing to %s : %s" % (self, statement))
            startswith = statement.split(None, 1)[0]
            # insert new statement with similar OR at the end of the file
            new_content_lines.insert(
                insert_locations.get(startswith, len(new_content_lines)),
                statement)

        if new_content_lines != original_content_lines:
            self.seek(0)
            self.writelines(new_content_lines)
            self.flush()
